from werkzeug.utils import validate_arguments, bind_arguments


def adv_validate_arguments(callback, args, kwargs):
    """
    "Advanced version" of Werkzeug's "validate_arguments" which doesn't modify
    passed args and kwargs

    :return: (args, kwargs) to pass to original function
    """
    bind_args = list(args)
    bind_kwargs = kwargs.copy()
    arguments, keyword_arguments = validate_arguments(callback, bind_args,
                                                      bind_kwargs)
    return arguments, keyword_arguments


def adv_bind_arguments(callback, args, kwargs):
    """
    "Advanced version" of Werkzeug's "bind_arguments" which doesn't modify
    passed args and kwargs
    """
    bind_args = list(args)
    bind_kwargs = kwargs.copy()
    keyword_arguments = bind_arguments(callback, bind_args, bind_kwargs)
    return keyword_arguments



import logging
import warnings

from sqs_workers.backoff_policies import DEFAULT_BACKOFF
from sqs_workers.context import SQSContext
from sqs_workers.utils import adv_validate_arguments
from sqs_workers import codecs

logger = logging.getLogger(__name__)

DEFAULT_CONTEXT_VAR = 'context'


class GenericProcessor(object):
    """
    Base superclass for all types of processors. Accepts queue name and job
    name to take care of, and optionally the processing function fn(...)

    The way job messages are actually processed, is defined in the
    process_batch() function which has to be re-defined in a subclass.

    Notes on the process_batch implementation.

    - job_messages of process_batch() is a list of SQS message objects
    http://boto3.readthedocs.io/en/latest/reference/services/sqs.html#message
    - The function must return two lists:
        - successfully processed messages
        - failed messages
    - The function is not supposed to die with any exception, instead the
      error has to be caught and logged, and the message has to be excluded
      from the list of successfully processed messages
    """

    def __init__(self,
                 sqs_env,
                 queue_name,
                 job_name,
                 fn=None,
                 pass_context=False,
                 context_var=DEFAULT_CONTEXT_VAR,
                 backoff_policy=DEFAULT_BACKOFF):
        self.sqs_env = sqs_env
        self.queue_name = queue_name
        self.job_name = job_name
        self.fn = fn
        self.pass_context = pass_context
        self.context_var = context_var
        self.backoff_policy = backoff_policy

    def __repr__(self):
        if self.fn:
            fn_name = self.fn.__module__ + '.' + self.fn.__name__
        else:
            fn_name = ''
        return '{}({!r}, {!r}, {})'.format(
            self.__class__.__name__, self.queue_name, self.job_name, fn_name)

    def process_batch(self, job_messages):
        raise NotImplementedError("Must be implemented in subclasses")

    def copy(self, **kwargs):
        """
        Create a new instance of the processor, optionally updating
        arguments of the constructor from update_kwargs
        """
        init_kwargs = {
            'sqs_env': kwargs.get('sqq_env', self.sqs_env),
            'queue_name': kwargs.get('queue_name', self.queue_name),
            'job_name': kwargs.get('job_name', self.job_name),
            'fn': kwargs.get('fn', self.fn),
            'pass_context': kwargs.get('pass_context', self.pass_context),
            'context_var': kwargs.get('context_var', self.context_var),
            'backoff_policy': kwargs.get('backoff_policy',
                                         self.backoff_policy),
        }
        return self.__class__(**init_kwargs)


class Processor(GenericProcessor):
    """
    Processor which calls its function for each incoming message.

    fn() accepts a single parameter "message" which is a decoded body of an
    SQS message
    """

    def process_batch(self, job_messages):
        succeeded, failed = [], []
        for message in job_messages:
            extra = {
                'message_id': message.message_id,
                'queue_name': self.queue_name,
                'job_name': self.job_name,
            }
            logger.debug(
                'Process {queue_name}.{job_name}'.format(**extra), extra=extra)

            try:
                content_type = get_job_content_type(message)
                extra['job_content_type'] = content_type
                codec = codecs.get_codec(content_type)
                job_kwargs = codec.deserialize(message.body)
                job_context = get_job_context(message, codec)
                self.process(job_kwargs, job_context)
            except Exception:
                logger.exception(
                    'Error while processing {queue_name}.{job_name}'.format(
                        **extra),
                    extra=extra)
                failed.append(message)
            else:
                succeeded.append(message)
        return succeeded, failed

    def process(self, job_kwargs, job_context):
        effective_kwargs = job_kwargs.copy()
        if self.pass_context:
            effective_kwargs[self.context_var] = job_context
        return call_handler(self.fn, effective_kwargs)


class BatchProcessor(GenericProcessor):
    """
    Processor which calls its function for the entire batch of incoming.

    The function is called for the list of decoded message bodies
    """

    def process_batch(self, job_messages):
        if job_messages:
            extra = {
                'job_count': len(job_messages),
                'queue_name': self.queue_name,
                'job_name': self.job_name,
            }
            logger.debug(
                'Process {job_count} messages '
                'to {queue_name}.{job_name}'.format(**extra),
                extra=extra)
            try:
                jobs, context = self.decode_messages(job_messages)
                self.process(jobs, context)
            except Exception:
                logger.exception(
                    'Error while processing {job_count} messages '
                    'to {queue_name}.{job_name}'.format(**extra),
                    extra=extra)
                return [], job_messages  # all failed
            else:
                return job_messages, []  # all succeeded

    def decode_messages(self, job_messages):
        jobs = []
        context = []
        for message in job_messages:
            content_type = get_job_content_type(message)
            codec = codecs.get_codec(content_type)
            job = codec.deserialize(message.body)
            jobs.append(job)
            context.append(get_job_context(message, codec))
        return jobs, context

    def process(self, jobs, context):
        if self.pass_context:
            return self.fn(jobs, **{self.context_var: context})
        return self.fn(jobs)


class FallbackProcessor(GenericProcessor):
    """
    Processor which is used by default when none of the processors is attached
    explicitly
    """

    def process_batch(self, job_messages):
        warnings.warn('Error while processing {}.{}'.format(
            self.queue_name, self.job_name))
        # return empty list for succeeded and failed messages
        # - the queue will not delete messages
        # - it will not put messages back to the queue either, so they
        #   automatically appear there on reaching the visibility timeout
        return [], []


class DeadLetterProcessor(GenericProcessor):
    """
    Generic processor which can be used for the dead-letter queue to push back
    messages to the main queue. Has opinion on how queues are organized
    and uses some hard-coded options.

    It is supposed to process queues "something_dead" which is supposed to be
    a configured dead-letter queue for "something". In case of FIFO queues
    names would be "something_dead.fifo" for "something.fifo".

    While processing the queue, the processor takes every message and push it
    back to the queue "something" (or something.fifo) with a hard-coded delay
    of 1 second.

    If the queue name does't end with "_dead" or "_dead.fifo", the
    DeadLetterProcessor behaves like generic FallbackProcessor: shows the error
    message and keep message in the queue. It's made to prevent from creating
    infinite loops when the message from the dead letter queue is pushed back
    to the same queue, then immediately processed by the same processor again,
    etc.

    Usage example:

    >>> env = SQSEnv(fallback_processor_maker=DeadLetterProcessor)
    >>> env.process_queue("foo_dead", shutdown_policy=IdleShutdown(10))

    This code takes all the messages in foo_dead queue and push them back to
    the foo queue. Then it waits 10 seconds to ensure no new messages appear,
    and quit.
    """

    def process_batch(self, job_messages):
        if not is_deadletter(self.queue_name):
            warnings.warn('Error while processing {}.{}'.format(
                self.queue_name, self.job_name))
            return [], []

        for message in job_messages:
            self.push_back_message(message)
        return job_messages, []  # all succeeded

    def push_back_message(self, message):
        queue_name = get_deadletter_upstream_name(self.queue_name)
        content_type = message.message_attributes['ContentType']['StringValue']
        job_context = message.message_attributes['JobContext']['StringValue']
        if queue_name.endswith('.fifo'):
            deduplication_id = message.attributes['MessageDeduplicationId']
            group_id = message.attributes['MessageGroupId']
            # FIFO queues don't allow to set delay_seconds
            delay_seconds = None
        else:
            deduplication_id = None
            group_id = None
            delay_seconds = 1
        logger.debug(
            'Push back dead letter job {}.{}'.format(self.queue_name,
                                                     self.job_name),
            extra={
                'target_queue_name': self.queue_name,
                'queue_name': self.queue_name,
                'job_name': self.job_name,
            })
        self.sqs_env.add_raw_job(queue_name, self.job_name, message.body,
                                 job_context, content_type, delay_seconds,
                                 deduplication_id, group_id)


def is_deadletter(queue_name):
    """
    Helper function which returns if queue has valid deadletter name
    (that is, ends with "_dead" or "_dead.fifo")
    """
    return queue_name.endswith('_dead') or queue_name.endswith('_dead.fifo')


def get_deadletter_upstream_name(queue_name):
    """
    Return upstream name for deadletter queue name: "foo" for "foo_dead"
    and "foo.fifo" for "foo_dead.fifo". Raise RuntimeError if queue doesn't
    have a valid deadletter name.
    """
    if queue_name.endswith('_dead'):
        return queue_name[:-len('_dead')]
    if queue_name.endswith('_dead.fifo'):
        return queue_name[:-len('_dead.fifo')] + '.fifo'
    raise RuntimeError("Not a deadletter queue")


def get_job_content_type(job_message):
    attrs = job_message.message_attributes or {}
    return (attrs.get('ContentType') or {}).get('StringValue')


def get_job_context(job_message, codec):
    attrs = job_message.message_attributes or {}
    serialized = (attrs.get('JobContext') or {}).get('StringValue')
    if not serialized:
        return SQSContext()
    deserialized = codec.deserialize(serialized)
    return SQSContext.from_dict(deserialized)


def call_handler(fn, kwargs):
    try:
        handler_args, handler_kwargs = adv_validate_arguments(fn, [], kwargs)
    except TypeError:
        # it may happen, if "fn" is not a function (but
        # a mock object, for example)
        handler_args, handler_kwargs = [], kwargs
    fn(*handler_args, **handler_kwargs)



import json
import logging
import multiprocessing
from itertools import groupby

import boto3

from sqs_workers import codecs, processors, context
from sqs_workers.backoff_policies import DEFAULT_BACKOFF
from sqs_workers.codecs import DEFAULT_CONTENT_TYPE
from sqs_workers.processor_mgr import ProcessorManager, ProcessorManagerProxy
from sqs_workers.shutdown_policies import NEVER_SHUTDOWN, NeverShutdown

logger = logging.getLogger(__name__)

DEFAULT_MESSAGE_GROUP_ID = 'default'


class BatchProcessingResult(object):
    succeeded = None
    failed = None

    def __init__(self, queue_name, succeeded=None, failed=None):
        self.queue_name = queue_name
        self.succeeded = succeeded or []
        self.failed = failed or []

    def update(self, succeeded, failed):
        self.succeeded += succeeded
        self.failed += failed

    def succeeded_count(self):
        return len(self.succeeded)

    def failed_count(self):
        return len(self.failed)

    def total_count(self):
        return self.succeeded_count() + self.failed_count()

    def __repr__(self):
        return '<BatchProcessingResult/%s/%s/%s>' % (
            self.queue_name, self.succeeded_count(), self.failed_count())


class SQSEnv(ProcessorManagerProxy):
    def __init__(self,
                 session=boto3,
                 queue_prefix='',
                 backoff_policy=DEFAULT_BACKOFF,
                 processor_maker=processors.Processor,
                 batch_processor_maker=processors.BatchProcessor,
                 fallback_processor_maker=processors.FallbackProcessor,
                 context_maker=context.SQSContext):
        """
        Initialize SQS environment with boto3 session
        """
        self.session = session
        self.sqs_client = session.client('sqs')
        self.sqs_resource = session.resource('sqs')
        self.queue_prefix = queue_prefix
        self.context = context_maker()
        self.processors = ProcessorManager(
            self, backoff_policy, processor_maker, batch_processor_maker,
            fallback_processor_maker)

        # internal mapping from queue names to queue objects
        self.queue_mapping_cache = {}

    def create_standard_queue(self,
                              queue_name,
                              message_retention_period=None,
                              visibility_timeout=None,
                              redrive_policy=None):
        """
        Create a new standard queue
        """
        attrs = {}
        kwargs = {
            'QueueName': self.get_sqs_queue_name(queue_name),
            'Attributes': attrs,
        }
        if message_retention_period is not None:
            attrs['MessageRetentionPeriod'] = str(message_retention_period)
        if visibility_timeout is not None:
            attrs['VisibilityTimeout'] = str(visibility_timeout)
        if redrive_policy is not None:
            attrs['RedrivePolicy'] = redrive_policy.__json__()
        ret = self.sqs_client.create_queue(**kwargs)
        return ret['QueueUrl']

    def create_fifo_queue(self,
                          queue_name,
                          content_based_deduplication=False,
                          message_retention_period=None,
                          visibility_timeout=None,
                          redrive_policy=None):
        """
        Create a new FIFO queue. Note that queue name has to end with ".fifo"

        - "content_based_deduplication" turns on automatic content-based
          deduplication of messages in the queue

        - redrive_policy can be None or an object, generated with
          redrive_policy() method of SQS. In the latter case if defines the
          way failed messages are processed.
        """
        attrs = {
            'FifoQueue': 'true',
        }
        kwargs = {
            'QueueName': self.get_sqs_queue_name(queue_name),
            'Attributes': attrs,
        }
        if content_based_deduplication:
            attrs['ContentBasedDeduplication'] = 'true'
        if message_retention_period is not None:
            attrs['MessageRetentionPeriod'] = str(message_retention_period)
        if visibility_timeout is not None:
            attrs['VisibilityTimeout'] = str(visibility_timeout)
        if redrive_policy is not None:
            attrs['RedrivePolicy'] = redrive_policy.__json__()
        ret = self.sqs_client.create_queue(**kwargs)
        return ret['QueueUrl']

    def purge_queue(self, queue_name):
        """
        Remove all messages from the queue
        """
        self.get_queue(queue_name).purge()

    def delete_queue(self, queue_name):
        """
        Delete the queue
        """
        self.get_queue(queue_name).delete()

    def add_job(self,
                queue_name,
                job_name,
                _content_type=DEFAULT_CONTENT_TYPE,
                _delay_seconds=None,
                _deduplication_id=None,
                _group_id=None,
                **job_kwargs):
        """
        Add job to the queue. The body of the job will be converted to the text
        with one of the codes (by default it's "pickle")
        """
        codec = codecs.get_codec(_content_type)
        message_body = codec.serialize(job_kwargs)
        job_context = codec.serialize(self.context.to_dict())
        return self.add_raw_job(queue_name, job_name, message_body,
                                job_context, _content_type, _delay_seconds,
                                _deduplication_id, _group_id)

    def add_raw_job(self, queue_name, job_name, message_body, job_context,
                    content_type, delay_seconds, deduplication_id, group_id):
        """
        Low-level function to put message to the queue
        """
        # if queue name ends with .fifo, then according to the AWS specs,
        # it's a FIFO queue, and requires group_id.
        # Otherwise group_id can be set to None
        if group_id is None and queue_name.endswith('.fifo'):
            group_id = DEFAULT_MESSAGE_GROUP_ID

        queue = self.get_queue(queue_name)
        kwargs = {
            'MessageBody': message_body,
            'MessageAttributes': {
                'ContentType': {
                    'StringValue': content_type,
                    'DataType': 'String',
                },
                'JobContext': {
                    'StringValue': job_context,
                    'DataType': 'String',
                },
                'JobName': {
                    'StringValue': job_name,
                    'DataType': 'String',
                },
            },
        }
        if delay_seconds is not None:
            kwargs['DelaySeconds'] = int(delay_seconds)
        if deduplication_id is not None:
            kwargs['MessageDeduplicationId'] = str(deduplication_id)
        if group_id is not None:
            kwargs['MessageGroupId'] = str(group_id)
        ret = queue.send_message(**kwargs)
        return ret['MessageId']

    def process_queues(self,
                       queue_names=None,
                       shutdown_policy_maker=NeverShutdown):
        """
        Use multiprocessing to process multiple queues at once. If queue names
        are not set, process all known queues

        shutdown_policy_maker is an optional callable which doesn't accept any
        arguments and create a new shutdown policy for each queue.

        Can looks somewhat like this:

            lambda: IdleShutdown(idle_seconds=10)
        """
        if not queue_names:
            queue_names = self.get_all_known_queues()
        processes = []
        for queue_name in queue_names:
            p = multiprocessing.Process(
                target=self.process_queue,
                kwargs={
                    'queue_name': queue_name,
                    'shutdown_policy': shutdown_policy_maker(),
                })
            p.start()
            processes.append(p)
        for p in processes:
            p.join()

    def drain_queue(self, queue_name, wait_seconds=0):
        """
        Delete all messages from the queue without calling purge()
        """
        queue = self.get_queue(queue_name)
        deleted_count = 0
        while True:
            messages = self.get_raw_messages(queue_name, wait_seconds)
            if not messages:
                break
            entries = [{
                'Id': msg.message_id,
                'ReceiptHandle': msg.receipt_handle
            } for msg in messages]
            queue.delete_messages(Entries=entries)
            deleted_count += len(messages)
        return deleted_count

    def process_queue(self,
                      queue_name,
                      shutdown_policy=NEVER_SHUTDOWN,
                      wait_second=10):
        """
        Run worker to process one queue in the infinite loop
        """
        logger.debug(
            'Start processing queue {}'.format(queue_name),
            extra={
                'queue_name': queue_name,
                'wait_seconds': wait_second,
                'shutdown_policy': repr(shutdown_policy),
            })
        while True:
            result = self.process_batch(queue_name, wait_seconds=wait_second)
            shutdown_policy.update_state(result)
            if shutdown_policy.need_shutdown():
                logger.debug(
                    'Stop processing queue {}'.format(queue_name),
                    extra={
                        'queue_name': queue_name,
                        'wait_seconds': wait_second,
                        'shutdown_policy': repr(shutdown_policy),
                    })
                break

    def process_batch(self, queue_name, wait_seconds=0):
        # type: (str, int) -> BatchProcessingResult
        """
        Process a batch of messages from the queue (10 messages at most), return
        the number of successfully processed messages, and exit
        """
        queue = self.get_queue(queue_name)
        messages = self.get_raw_messages(queue_name, wait_seconds)
        result = BatchProcessingResult(queue_name)

        for job_name, job_messages in group_messages(queue_name, messages):
            processor = self.processors.get(queue_name, job_name)
            succeeded, failed = processor.process_batch(job_messages)
            result.update(succeeded, failed)
            if succeeded:
                entries = [{
                    'Id': msg.message_id,
                    'ReceiptHandle': msg.receipt_handle
                } for msg in succeeded]
                queue.delete_messages(Entries=entries)
            if failed:
                entries = [{
                    'Id':
                    msg.message_id,
                    'ReceiptHandle':
                    msg.receipt_handle,
                    'VisibilityTimeout':
                    processor.backoff_policy.get_visibility_timeout(msg),
                } for msg in failed]
                queue.change_message_visibility_batch(Entries=entries)
        return result

    def get_raw_messages(self, queue_name, wait_seconds):
        """
        Return raw messages from the queue, addressed by its name
        """
        kwargs = {
            'WaitTimeSeconds': wait_seconds,
            'MaxNumberOfMessages': 10,
            'MessageAttributeNames': ['All'],
            'AttributeNames': ['All'],
        }
        queue = self.get_queue(queue_name)
        return queue.receive_messages(**kwargs)

    def get_queue(self, queue_name):
        """
        Helper function to return queue object by name
        """
        if queue_name not in self.queue_mapping_cache:
            queue = self.sqs_resource.get_queue_by_name(
                QueueName=self.get_sqs_queue_name(queue_name))
            self.queue_mapping_cache[queue_name] = queue
        return self.queue_mapping_cache[queue_name]

    def get_all_known_queues(self):
        resp = self.sqs_client.list_queues(
            **{
                'QueueNamePrefix': self.queue_prefix,
            })
        if 'QueueUrls' not in resp:
            return []
        urls = resp['QueueUrls']
        ret = []
        for url in urls:
            sqs_name = url.rsplit('/', 1)[-1]
            ret.append(sqs_name[len(self.queue_prefix):])
        return ret

    def get_sqs_queue_name(self, queue_name):
        """
        Take "high-level" (user-visible) queue name and return SQS
        ("low level") name by simply prefixing it. Used to create namespaces
        for different environments (development, staging, production, etc)
        """
        return '{}{}'.format(self.queue_prefix, queue_name)

    def redrive_policy(self, dead_letter_queue_name, max_receive_count):
        return RedrivePolicy(self, dead_letter_queue_name, max_receive_count)


class RedrivePolicy(object):
    """
    Redrive Policy for SQS queues

    See for more details:
    https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-dead-letter-queues.html
    """

    def __init__(self, sqs_env, dead_letter_queue_name, max_receive_count):
        self.sqs_env = sqs_env
        self.dead_letter_queue_name = dead_letter_queue_name
        self.max_receive_count = max_receive_count

    def __json__(self):
        queue = self.sqs_env.sqs_resource.get_queue_by_name(
            QueueName=self.sqs_env.get_sqs_queue_name(
                self.dead_letter_queue_name))
        target_arn = queue.attributes['QueueArn']
        # Yes, it's double-encoded JSON :-/
        return json.dumps({
            'deadLetterTargetArn': target_arn,
            'maxReceiveCount': str(self.max_receive_count),
        })


def group_messages(queue_name, messages):
    """
    Group SQSMessage instances by JobName attribute. If queue name ends with
    .fifo (it's a FIFO queue), we don't reorder messages while grouping,
    otherwise messages can get out of the grouper reorderd
    """
    if not queue_name.endswith('.fifo'):
        messages = sorted(messages, key=get_job_name)
    for job_name, job_messages in groupby(messages, key=get_job_name):
        yield job_name, list(job_messages)


def get_job_name(message):
    attrs = message.message_attributes or {}
    return (attrs.get('JobName') or {}).get('StringValue')



import logging
import warnings
from collections import defaultdict

from sqs_workers import processors
from sqs_workers.backoff_policies import DEFAULT_BACKOFF
from sqs_workers.codecs import DEFAULT_CONTENT_TYPE
from sqs_workers.processors import DEFAULT_CONTEXT_VAR
from sqs_workers.utils import adv_bind_arguments

logger = logging.getLogger(__name__)


class ProcessorManagerProxy(object):

    processors = None  # type: "ProcessorManager"

    def processor(self,
                  queue_name,
                  job_name,
                  pass_context=False,
                  context_var=DEFAULT_CONTEXT_VAR,
                  backoff_policy=None):
        """
        Decorator to assign processor to handle jobs with the name job_name
        from the queue queue_name

        Usage example:


        @sqs.processor('q1', 'say_hello')
        def say_hello(name):
            print("Hello, " + name)

        Then you can add messages to the queue by calling ".delay" attribute
        of a newly created object:

        >>> say_hello.delay(name='John')

        Here, instead of calling function locally, it will be pushed to the
        queue (essentially, mimics the non-blocking call of the function).

        If you still want to call function locally, you can call

        >>> say_hello(name='John')
        """

        def fn(processor):
            return self.processors.connect(queue_name, job_name, processor,
                                           pass_context, context_var,
                                           backoff_policy)

        return fn

    def batch_processor(self,
                        queue_name,
                        job_name,
                        pass_context=False,
                        context_var=DEFAULT_CONTEXT_VAR,
                        backoff_policy=None):
        """
        Decorator to assign a batch processor to handle jobs with the name
        job_name from the queue queue_name. The batch processor has to be a
        function which accepts the only argument "jobs".

        "jobs" is a list of properly decoded dicts

        Usage example:


        @sqs.processors.connect('q1', 'batch_say_hello')
        def batch_say_hello(jobs):
            names = ', '.join([job['name'] for job in jobs])
            print("Hello, " + names)

        Then you can add messages to the queue by calling ".delay" attribute
        of a newly created object:

        >>> batch_say_hello.delay(name='John')

        Here, instead of calling function locally, it will be pushed to the
        queue (essentially, mimics the non-blocking call of the function).

        If you still want to call function locally, you can call

        >>> batch_say_hello(name='John')
        """

        def fn(processor):
            return self.processors.connect_batch(queue_name, job_name,
                                                 processor, pass_context,
                                                 context_var, backoff_policy)

        return fn


class ProcessorManager(object):
    def __init__(
            self,
            sqs_env,
            backoff_policy=DEFAULT_BACKOFF,
            processor_maker=processors.Processor,
            batch_processor_maker=processors.BatchProcessor,
            fallback_processor_maker=processors.FallbackProcessor,
    ):
        self.sqs_env = sqs_env
        self.processors = defaultdict(lambda: {})
        self.backoff_policy = backoff_policy
        self.processor_maker = processor_maker
        self.batch_processor_maker = batch_processor_maker
        self.fallback_processor_maker = fallback_processor_maker
        self.processors = defaultdict(lambda: {})

    def connect(self,
                queue_name,
                job_name,
                processor,
                pass_context=False,
                context_var=DEFAULT_CONTEXT_VAR,
                backoff_policy=None):
        """
        Assign processor (a function) to handle jobs with the name job_name
        from the queue queue_name
        """
        extra = {
            'queue_name': queue_name,
            'job_name': job_name,
            'processor_name': processor.__module__ + '.' + processor.__name__,
        }
        logger.debug(
            'Connect {queue_name}.{job_name} to '
            'processor {processor_name}'.format(**extra),
            extra=extra)
        self.processors[queue_name][job_name] = self.processor_maker(
            self.sqs_env, queue_name, job_name, processor, pass_context, context_var,
            backoff_policy or self.backoff_policy)
        return AsyncTask(self.sqs_env, queue_name, job_name, processor)

    def connect_batch(self,
                      queue_name,
                      job_name,
                      processor,
                      pass_context=False,
                      context_var=DEFAULT_CONTEXT_VAR,
                      backoff_policy=None):
        """
        Assign a batch processor (function) to handle jobs with the name
        job_name from the queue queue_name
        """
        extra = {
            'queue_name': queue_name,
            'job_name': job_name,
            'processor_name': processor.__module__ + '.' + processor.__name__,
        }
        logger.debug(
            'Connect {queue_name}.{job_name} to '
            'batch processor {processor_name}'.format(**extra),
            extra=extra)
        self.processors[queue_name][job_name] = self.batch_processor_maker(
            self.sqs_env, queue_name, job_name, processor, pass_context, context_var,
            backoff_policy or self.backoff_policy)
        return AsyncBatchTask(self.sqs_env, queue_name, job_name, processor)

    def copy(self, src_queue, dst_queue):
        """
        Copy processors from src_queue to dst_queue (both queues identified
        by their names). Can be helpful to process dead-letter queue with
        processors from the main queue.

        Usage example.

        sqs = SQSEnv()
        ...
        sqs.processors.copy('foo', 'foo_dead')
        sqs.process_queue("foo_dead", shutdown_policy=IdleShutdown(10))

        Here the queue "foo_dead" will be processed with processors from the
        queue "foo".
        """
        for job_name, processor in self.processors[src_queue].items():
            self.processors[dst_queue][job_name] = processor.copy(
                queue_name=dst_queue)

    def get(self, queue_name, job_name):
        """
        Helper function to return a processor for the queue
        """
        processor = self.processors[queue_name].get(job_name)
        if processor is None:
            processor = self.fallback_processor_maker(self.sqs_env, queue_name,
                                                      job_name)
            self.processors[queue_name][job_name] = processor
        return processor


class AsyncTask(object):
    def __init__(self, sqs_env, queue_name, job_name, processor):
        self.sqs_env = sqs_env
        self.queue_name = queue_name
        self.job_name = job_name
        self.processor = processor
        self.__doc__ = processor.__doc__

    def __call__(self, *args, **kwargs):
        warnings.warn(
            'Async task {0.queue_name}.{0.job_name} called synchronously'.
            format(self))
        return self.processor(*args, **kwargs)

    def __repr__(self):
        return '<%s %s.%s>' % (self.__class__.__name__, self.queue_name,
                               self.job_name)

    def delay(self, *args, **kwargs):
        _content_type = kwargs.pop('_content_type', DEFAULT_CONTENT_TYPE)
        _delay_seconds = kwargs.pop('_delay_seconds', None)
        _deduplication_id = kwargs.pop('_deduplication_id', None)
        _group_id = kwargs.pop('_group_id', None)
        kwargs = adv_bind_arguments(self.processor, args, kwargs)
        return self.sqs_env.add_job(
            self.queue_name,
            self.job_name,
            _content_type=_content_type,
            _delay_seconds=_delay_seconds,
            _deduplication_id=_deduplication_id,
            _group_id=_group_id,
            **kwargs)


class AsyncBatchTask(AsyncTask):
    def __call__(self, **kwargs):
        warnings.warn(
            'Async task {0.queue_name}.{0.job_name} called synchronously'.
            format(self))
        return self.processor([kwargs])

    def delay(self, **kwargs):
        # unlike AsyncTask, we cannot validate or bind arguments, because
        # processor always accepts the list of dicts, so we just pass kwargs
        # as is
        return self.sqs_env.add_job(self.queue_name, self.job_name, **kwargs)



import datetime
import logging
import multiprocessing
import time
import uuid
from queue import Empty, Queue

from sqs_workers import codecs, context, processors
from sqs_workers.backoff_policies import DEFAULT_BACKOFF
from sqs_workers.core import (
    DEFAULT_CONTENT_TYPE, BatchProcessingResult, RedrivePolicy, group_messages)
from sqs_workers.processor_mgr import ProcessorManager, ProcessorManagerProxy
from sqs_workers.shutdown_policies import NEVER_SHUTDOWN, NeverShutdown

logger = logging.getLogger(__name__)


class MemoryEnv(ProcessorManagerProxy):
    """
    In-memory pseudo SQS implementation, for faster and more predictable
    processing in tests.

    Implements the SQSEnv interface, but lacks some features of it, and some
    other features implemented ineffectively.

    - Redrive policy doesn't work
    - There is no differences between standard and FIFO queues
    - FIFO queues don't support content-based deduplication
    - Delayed tasks executed ineffectively: the task is gotten from the queue,
      and if the time hasn't come yet, the task is put back.
    - API can return slightly different results
    """

    # TODO: maybe it can be implemented more effectively with sched

    def __init__(self,
                 backoff_policy=DEFAULT_BACKOFF,
                 processor_maker=processors.Processor,
                 batch_processor_maker=processors.BatchProcessor,
                 fallback_processor_maker=processors.FallbackProcessor,
                 context_maker=context.SQSContext):
        """
        Initialize pseudo SQS environment
        """
        self.backoff_policy = backoff_policy
        self.context = context_maker()
        self.processors = ProcessorManager(
            self, backoff_policy, processor_maker, batch_processor_maker,
            fallback_processor_maker)
        self.queues = {}  # type: dict[str, Queue]

    def create_standard_queue(self, queue_name, **kwargs):
        """
        Create a new standard queue
        """
        self.queues[queue_name] = Queue()
        return 'memory://%s' % queue_name

    def create_fifo_queue(self, queue_name, **kwargs):
        """
        Create a new FIFO queue. Implementation is the same as for the standard
        queue
        """
        self.queues[queue_name] = Queue()
        return 'memory://%s' % queue_name

    def purge_queue(self, queue_name):
        """
        Remove all messages from the queue
        """
        queue = self.queues[queue_name]
        while True:
            try:
                queue.get_nowait()
            except Empty:
                return

    def delete_queue(self, queue_name):
        """
        Delete the queue
        """
        self.queues.pop(queue_name)

    def add_job(self,
                queue_name,
                job_name,
                _content_type=DEFAULT_CONTENT_TYPE,
                _delay_seconds=None,
                _deduplication_id=None,
                _group_id=None,
                **job_kwargs):
        """
        Add job to the queue. The body of the job will be converted to the text
        with one of the codes (by default it's "pickle")
        """
        codec = codecs.get_codec(_content_type)
        message_body = codec.serialize(job_kwargs)
        job_context = codec.serialize(self.context.to_dict())
        return self.add_raw_job(queue_name, job_name, message_body,
                                job_context, _content_type, _delay_seconds,
                                _deduplication_id, _group_id)

    def add_raw_job(self, queue_name, job_name, message_body, job_context,
                    content_type, delay_seconds, deduplication_id, group_id):
        """
        Low-level function to put message to the queue
        """
        kwargs = {
            'MessageBody': message_body,
            'MessageAttributes': {
                'ContentType': {
                    'StringValue': content_type,
                    'DataType': 'String',
                },
                'JobContext': {
                    'StringValue': job_context,
                    'DataType': 'String',
                },
                'JobName': {
                    'StringValue': job_name,
                    'DataType': 'String',
                },
            },
        }
        if delay_seconds is not None:
            delay_seconds_int = int(delay_seconds)
            execute_at = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=delay_seconds)
            kwargs['DelaySeconds'] = delay_seconds_int
            kwargs['_execute_at'] = execute_at
        queue = self.queues[queue_name]
        queue.put(kwargs)
        return ''

    def process_queues(self,
                       queue_names=None,
                       shutdown_policy_maker=NeverShutdown):
        """
        Use multiprocessing to process multiple queues at once. If queue names
        are not set, process all known queues

        shutdown_policy_maker is an optional callable which doesn't accept any
        arguments and create a new shutdown policy for each queue.
        """
        if not queue_names:
            queue_names = self.get_all_known_queues()
        processes = []
        for queue_name in queue_names:
            p = multiprocessing.Process(
                target=self.process_queue,
                kwargs={
                    'queue_name': queue_name,
                    'shutdown_policy': shutdown_policy_maker(),
                })
            p.start()
            processes.append(p)
        for p in processes:
            p.join()

    def drain_queue(self, queue_name, wait_seconds=0):
        """
        Delete all messages from the queue. An equivalent to purge()
        """
        self.purge_queue(queue_name)

    def process_queue(self,
                      queue_name,
                      shutdown_policy=NEVER_SHUTDOWN,
                      wait_second=10):
        """
        Run worker to process one queue in the infinite loop
        """
        while True:
            result = self.process_batch(queue_name, wait_seconds=wait_second)
            shutdown_policy.update_state(result)
            if shutdown_policy.need_shutdown():
                break

    def process_batch(self, queue_name, wait_seconds=0):
        # type: (str, int) -> BatchProcessingResult
        """
        Process a batch of messages from the queue (10 messages at most), return
        the number of successfully processed messages, and exit
        """
        queue = self.queues[queue_name]
        messages = self.get_raw_messages(queue_name, wait_seconds)
        result = BatchProcessingResult(queue_name)
        for job_name, job_messages in group_messages(queue_name, messages):
            processor = self.processors.get(queue_name, job_name)
            succeeded, failed = processor.process_batch(job_messages)
            result.update(succeeded, failed)
            if failed:
                for msg in failed:
                    queue.put(msg)
        return result

    def get_raw_messages(self, queue_name, wait_seconds=0):
        """
        Helper function to get at most 10 messages from the queue, waiting for
        wait_seconds at most before they get ready.
        """
        max_messages = 10
        sleep_interval = 0.1

        messages = []
        while True:
            messages += self._get_some_raw_messages(queue_name, max_messages)
            if len(messages) >= max_messages:
                break
            if wait_seconds <= 0:
                break
            wait_seconds -= sleep_interval
            time.sleep(sleep_interval)

        return messages

    def _get_some_raw_messages(self, queue_name, max_messages):
        """
        Helper function which returns at most max_messages from the
        queue. Used in an infinite loop inside `get_raw_messages`
        """
        queue = self.queues[queue_name]
        messages = []
        for i in range(max_messages):
            try:
                messages.append(RawMessage(queue.get_nowait()))
            except Empty:
                break

        now = datetime.datetime.utcnow()
        ready_messages = []
        for message in messages:
            execute_at = message.get('_execute_at', now)
            if execute_at > now:
                queue.put(message)
            else:
                ready_messages.append(message)

        return ready_messages

    def get_all_known_queues(self):
        return list(self.queues.keys())

    def get_sqs_queue_name(self, queue_name):
        return queue_name

    def redrive_policy(self, dead_letter_queue_name, max_receive_count):
        return RedrivePolicy(self, dead_letter_queue_name, max_receive_count)


class RawMessage(dict):
    """
    A mock class to mimic the AWS message
    """

    @property
    def message_attributes(self):
        return self['MessageAttributes']

    @property
    def body(self):
        return self['MessageBody']

    @property
    def message_id(self):
        return uuid.uuid4().hex
